/**
 * File:
 *  modbusSlave.c
 *
 * Notes:
 *  This file contains the implementation for a serial RTU modbus slave.
 *
 * Functions:
 *  addCoil         Add a coil to the slave
 *  addHoldingReg   Add a holding register to the slave
 *  addInputReg     Add an input register to the slave
 *  addStatusBit    Add a status bit to the slave
 *  decodePacket    Decodes a packet received from modbus master
 *  modbusSlave     Initialise modbus slave structure
 *
 * History:
 *  10/06/2012 Written by Simon Platten
 *  19/06/2012 Added receivePacket function which buffers a packet
 *  21/06/2012 Implemented exception responses
 *             Implemented Timer0 to time receipt of bytes
 *  25/06/2012 Implemented startTimeout
 */
#include "modbusSlave.h"
// Pointers to last node in each linked list
static coilDef*    pLastCoil       = NULL;
static statusDef*  pLastStatusBit  = NULL;
static inputDef*   pLastInput      = NULL;
static holdingDef* pLastHoldingReg = NULL;
// Pointers to linked lists for each data type
static coilDef*    pCoils          = NULL;
static statusDef*  pStatusBits     = NULL;
static inputDef*   pInputs         = NULL;
static holdingDef* pHoldingRegs    = NULL;
// Flags to use when validating address ranges
static boolean blnValidStart, blnValidEnd;
// Exception code
static mbException exceptionCode;
// The modbus slave address
static byte bytSlaveAddress;
// The requested start address high byte and low byte
static byte bytStartHi, bytStartLo;
// The requested total number of items high byte and low byte
static byte bytTotalHi, bytTotalLo;
// The byte index
static byte bytIndex = 0;
// Receiver GAP set-point
static uint uintRxGapSetPt;
// Receiver GAP counter
static uint uintRxGap;
// Flag to indicate when to check buffer
static boolean blnCheckBuffer = FALSE;
// Received packet buffer
union {
  byte          arybytes[MAX_RECEIVED_PACKET_LENGTH];
  reqCoilsDef   reqCoils;
  reqStatusDef  reqStatus;
  reqHoldReqDef reqHoldReg;
  reqInputsDef  reqInputs;
} request;

union {
  byte          arybytes[MAX_RESPONSE_PACKET_LENGTH];
} reply;
/**
 * Function:
 *  addDiscrete
 *
 * Parameters:
 *  pDiscrete, pointer to an instance of discreteDef
 *  uintAddress, the address to assign to the discrete (base 1)
 *  pByte, pointer to the byte containing the bit to map as the discrete
 *  bBit, the bit in the byte to map to the discrete, 0 to 7
 *  pLinkList, pointer to pointer of discrete linked list
 *  pLast, pointer to pointer of last node in list
 *
 * Returns:
 *  TRUE if discrete added, FALSE if not
 */
static boolean addDiscrete(discreteDef* pDiscrete,
                           uint uintAddress,
                           ushort* pByte, byte bBit,
                           discreteDef** pLinkList,
                           discreteDef** pLast) {
  if ( !(pDiscrete && uintAddress && pByte &&
         bBit >= 0 && bBit < 8 && pLinkList && pLast) ) {
    return FALSE;
  }
  if ( *pLinkList == NULL ) {
// This is the first coil
    *pLinkList = *pLast = pDiscrete;
  } else {
// Append this coil to the linked list
    (*pLast)->pNext = pDiscrete;
// Adjust the last node
    *pLast = pDiscrete;
  }
// Ensure the discrete is set-up using the passed parameters
  pDiscrete->uintAddress = uintAddress;
  pDiscrete->pByte = pByte;
  pDiscrete->bBit  = bBit;
  pDiscrete->pNext = NULL;
}
/**
 * Function:
 *  addRegister
 *
 * Parameters:
 *  pRegister, pointer to an instance of registerDef
 *  uintAddress, the address to assign to the register (base 1)
 *  pWord, pointer to the word to map as the register
 *  pLinkList, pointer to pointer of register linked list
 *  pLast, pointer to pointer of last node in list
 *
 * Returns:
 *  TRUE if register added, FALSE if not
 */
static boolean addRegister(registerDef* pRegister,
                           uint uintAddress,
                           ushort* pWord,
                           registerDef** pLinkList,
                           registerDef** pLast) {
  if ( !(pRegister && uintAddress && pWord && pLinkList && pLast) ) {
    return FALSE;
  }
  if ( *pLinkList == NULL ) {
// This is the first coil
    *pLinkList = *pLast = pRegister;
  } else {
// Append this coil to the linked list
    (*pLast)->pNext = pRegister;
// Adjust the last node
    *pLast = pRegister;
  }
// Ensure the discrete is set-up using the passed parameters
  pRegister->uintAddress = uintAddress;
  pRegister->pWord = pWord;
  pRegister->pNext = NULL;
}
/**
 * Function:
 *  calcCRC
 *
 * Parameters:
 *  pMsg, the message containing the bytes to calculate the CRC for
 *  uintLength, the number of bytes in the message
 *
 * Returns:
 *  A 16bit CRC for the passed message
 */
static long calcCRC(byte* pMsg, uint uintLength) {
  long lngCRC = 0xffff;
  uint i, j;
  for( i=0; i<uintLength; i++ ) {
    lngCRC ^= (pMsg[i] & 0xff);

    for( j=0; j<8; j++ ) {
      if ( lngCRC & 0x01 ) {
        lngCRC >>= 0x01;
        lngCRC ^= 0xa001;
      } else {
        lngCRC >>= 0x01;
      }
    }
  }
  return lngCRC;
}
/**
 * Function:
 *  coilState
 *
 * Parameters:
 *  uintAddr, the address of the coil to modify
 *  eState, TRUE to turn coil on, FALSE to turn coil off
 *
 * Returns:
 *  none
 */
static void coilState(uint uintAddr, boolean eState) {
// Find the first I/O point
  discreteDef* pNode;
  byte bytMask;
  for( pNode=pCoils; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress != uintAddr  ) {
      continue;
    }
    bytMask = (1 << pNode->bBit);

    if ( eState == TRUE ) {
// Turn on
      *pNode->pByte |= bytMask;
    } else {
// Turn off
      *pNode->pByte &= (0xffff ^ bytMask);
    }
    break;
  }
}
/**
 * Function:
 *  packBits
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the discrete items
 *  pBuffer, a pointer to the buffer to pack the bits into
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  The number of bytes the bits were packed into
 */
static uint packBits(discreteDef* pLL, byte* pBuffer, 
                     uint uintStart, uint uintEnd) {
// Find the first I/O point
  uint uintBytes = 0, uintBit = 0;
  discreteDef* pNode;
  byte bytSet;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress >= uintStart && pNode->uintAddress <= uintEnd ) {
      bytSet = (*pNode->pByte & (1 << pNode->bBit)) ? 1 : 0;
      
      if ( bytSet ) {
        pBuffer[uintBytes] |= (bytSet << uintBit);
      }
      if ( ++uintBit > 7 ) {
        uintBit = 0;
        uintBytes++;
      }
    }
  }
  return uintBytes + 1;
}
/**
 * Function:
 *  packRegisters
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the register items
 *  pBuffer, a pointer to the buffer to pack the bits into
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  The number of bytes the registers were packed into
 */
static uint packRegisters(registerDef* pLL, byte* pBuffer,
                          uint uintStart, uint uintEnd) {
// Find the first I/O point
  uint uintBytes = 0;
  registerDef* pNode;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress >= uintStart && pNode->uintAddress <= uintEnd ) {
      pBuffer[uintBytes++] = (*pNode->pWord >> 0x08);
      pBuffer[uintBytes++] = (*pNode->pWord & 0xff);
    }
  }
  return uintBytes;
}
/**
 * Function:
 *  setRegister
 *
 * Parameters:
 *  uintAddr, the address of the register to modify
 *  bytHi, the hi-byte of the data value
 *  bytLo, the lo-byte of the data value
 *
 * Returns:
 *  none
 */
static void setRegister(uint uintAddr, byte bytHi, byte bytLo) {
// Find the first I/O point
  registerDef* pNode;
  for( pNode=pHoldingRegs; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress != uintAddr  ) {
      continue;
    }
    *pNode->pWord = ((int)bytHi << 0x08) | (int)bytLo;
    break;
  }
}
/**
 * Function:
 *  startTimeout
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
static void startTimeout(void) {
// Disable interrupt
  TMR0IE_bit = 0;
// Set timer presets
  TMR0H = TIMER0_PRESET_MSB;
  TMR0L = TIMER0_PRESET_LSB;
// Reset GAP counter
  uintRxGap = 0;
// Clear the interrupt mask
  TMR0IF_bit = 0;
// Enable interrupt
  TMR0IE_bit = 1;
}
/**
 * Function:
 *  validAddresses
 *
 * Parameters:
 *  uintAddress, the address to validate against
 *  uintStart, the first address to validate
 *  uintEnd, the last address to validate
 *
 * Returns:
 *  TRUE of FALSE if both addresses are valid
 */
static boolean validAddresses(uint uintAddress, uint uintStart, uint uintEnd) {
  if ( blnValidStart == FALSE && uintStart == uintAddress ) {
    blnValidStart = TRUE;
  }
  if ( blnValidEnd == FALSE && uintEnd == uintAddress ) {
    blnValidEnd = TRUE;
  }
  return (blnValidStart == TRUE && blnValidEnd == TRUE);
}
/**
 * Function:
 *  validDiscretes
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the discrete items
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  TRUE if the start and end addresses are valid, FALSE if not
 */
static boolean validDiscretes(discreteDef* pLL, uint uintStart, uint uintEnd) {
  discreteDef* pNode;
  blnValidStart = blnValidEnd = FALSE;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( validAddresses(pNode->uintAddress, uintStart, uintEnd) == TRUE ) {
      return TRUE;
    }
  }
  return FALSE;
}
/**
 * Function:
 *  validRegisters
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the register items
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  TRUE if the start and end addresses are valid, FALSE if not
 */
static boolean validRegisters(registerDef* pLL, uint uintStart, uint uintEnd) {
  registerDef* pNode;
  blnValidStart = blnValidEnd = FALSE;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( validAddresses(pNode->uintAddress, uintStart, uintEnd) == TRUE ) {
      return TRUE;
    }
  }
  return FALSE;
}
/**
 * Function:
 *  addCoil
 *
 * Parameters:
 *  pCoil, pointer to an instance of coilDef
 *  uintAddress, the address to assign to the coil (base 1)
 *  pByte, pointer to the byte containing the bit to map as the coil
 *  bBit, the bit in the byte to map to the coil, 0 to 7
 *
 * Returns:
 *  TRUE if coil added, FALSE if not
 */
boolean addCoil(coilDef* pCoil, uint uintAddress, ushort* pByte, byte bBit) {
  return addDiscrete(pCoil, uintAddress, pByte, bBit, &pCoils, &pLastCoil);
}
/**
 * Function:
 *  addHoldingReg
 *
 * Parameters:
 *  pHolding, pointer to an instance of holdingDef
 *  uintAddress, the address to assign to the holding register (base 1)
 *  pWord, pointer to the word to use for the holding register
 *
 * Returns:
 *  TRUE if coil added, FALSE if not
 */
boolean addHoldingReg(holdingDef* pHolding, uint uintAddress, ushort* pWord) {
  return addRegister(pHolding, uintAddress, pWord, &pHoldingRegs, 
                     &pLastHoldingReg);
}
/**
 * Function:
 *  addInputReg
 *
 * Parameters:
 *  pInput, pointer to an instance of inputDef
 *  uintAddress, the address to assign to the input register (base 1)
 *  pWord, pointer to the word to use for the input register
 *
 * Returns:
 *  TRUE if coil added, FALSE if not
 */
boolean addInputReg(inputDef* pInput, uint uintAddress, ushort* pWord) {
  return addRegister(pInput, uintAddress, pWord, &pInputs, &pLastInput);
}
/**
 * Function:
 *  addStatusBit
 *
 * Parameters:
 *  pStatus, pointer to an instance of statusDef
 *  uintAddress, the address to assign to the status bit (base 1)
 *  pByte, pointer to the byte containing the bit to map as the status bit
 *  bBit, the bit in the byte to map to the status bit, 0 to 7
 *
 * Returns:
 *  TRUE if status bit added, FALSE if not
 */
boolean addStatusBit(statusDef* pStatus, uint uintAddress, ushort* pByte, 
                     byte bBit) {
  return addDiscrete(pStatus, uintAddress, pByte, bBit, &pStatusBits, 
                     &pLastStatusBit);
}
/**
 * Function:
 *  decodePacket
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
void decodePacket(void) {
  uint uintCRCpos = 0;
  long lngCRC;

  if ( TMR0IF_bit == 1 ) {
    if ( blnCheckBuffer == FALSE && ++uintRxGap >= uintRxGapSetPt ) {
      if ( bytIndex >= 8 ) {
// The absolute smallest modbus packet consists of the 
// slave address, function code, start address, number of bytes and CRC
// Flag that its ok to examine the RX buffer now
        blnCheckBuffer = TRUE;
      }
// Gap of 3.5 bytes detected, anything to decode?
      uintRxGap = 0;
    }
    TMR0H = TIMER0_PRESET_MSB;
    TMR0L = TIMER0_PRESET_LSB;
// Clear the interrupt mask
    TMR0IF_bit = 0;
  }
  if( RCIF_bit ) {
    if ( bytIndex >= MAX_RECEIVED_PACKET_LENGTH ) {
// Prevent the index from overflowing the buffer
      bytIndex = 0;
    }
// Store the data in the buffer
    request.arybytes[bytIndex++] = (byte)UART_Read();
// Start / Reset timeout
    startTimeout();
// Clear the interrupt mask
    RCIF_bit = 0;
  }
  if ( blnCheckBuffer == FALSE ) {
// Nothing more to do, not read to process buffer yet!
    return;
  }
// Is this message address to this slave?
  if ( request.arybytes[0] != bytSlaveAddress ) {
// No, don't bother progressing
    blnCheckBuffer = FALSE;
    return;
  }
// Set the expected position of the CRC
  uintCRCpos = 6;
// Is the function supported?
  switch( request.arybytes[1] ) {
  case READ_COILS:
  case FORCE_SINGLE_COIL:
  case FORCE_MULTIPLE_COILS:
    if ( pCoils == NULL ) {
// No coils defined!
      exceptionCode = ILLEGAL_DATA_ADDRESS;
    }
    break;
  case READ_STATUS_INPUTS:
    if ( pStatusBits == NULL ) {
// No status inputs defined!
      exceptionCode = ILLEGAL_DATA_ADDRESS;
    }
    break;
  case READ_HOLDING_REGISTERS:
  case PRESET_SINGLE_REGISTER:
  case PRESET_MULTIPLE_REGISTERS:
    if ( pHoldingRegs == NULL ) {
// No holding registers defined!
      exceptionCode = ILLEGAL_DATA_ADDRESS;
    }
    break;
  case READ_INPUT_REGISTERS:
    if ( pInputs == NULL ) {
// No input registers defined!
      exceptionCode = ILLEGAL_DATA_ADDRESS;
    }
    break;
  default:
// No, function not supported!
    exceptionCode = ILLEGAL_FUNCTION;
    break;
  }
  bytStartHi = request.arybytes[2];
  bytStartLo = request.arybytes[3];
  bytTotalHi = request.arybytes[4];
  bytTotalLo = request.arybytes[5];

  if ( request.arybytes[1] == FORCE_MULTIPLE_COILS ||
       request.arybytes[1] == PRESET_MULTIPLE_REGISTERS )  {
    uintCRCpos += 1 + request.arybytes[6];
  }
// Calculate the packet CRC
  lngCRC = calcCRC(request.arybytes, uintCRCpos);
// Compare the calculate checksum with the received checksum
  if ( (lngCRC & 0xff)  == request.arybytes[uintCRCpos] &&
       (lngCRC >> 0x08) == request.arybytes[uintCRCpos + 1] ) {
// CRC ok
    uint uintSAddr, uintEAddr, uintItemCount, uintLength = 6;
    uint uintOffset, uintAddr, uintBit;
    discreteDef* pDiscretes;
    registerDef* pRegisters;
// Get the start address
    uintSAddr = (((uint)bytStartHi << 0x08) | bytStartLo) + 1;
// How many items have been requested?
    uintItemCount = ((uint)bytTotalHi << 0x08) | bytTotalLo;
// Whats the last address?
    uintEAddr = uintSAddr + uintItemCount - 1;
// Insert slave address and function code into response
    reply.arybytes[0] = request.arybytes[0];
    reply.arybytes[1] = request.arybytes[1];
// What was the requested function code?
    switch( request.arybytes[1] ) {
    case READ_COILS:
    case READ_STATUS_INPUTS:
      if ( request.arybytes[1] == READ_COILS ) {
        pDiscretes = pCoils;
      } else {
        pDiscretes = pStatusBits;
      }
      if ( validDiscretes(pDiscretes, uintSAddr, uintEAddr) == FALSE ) {
        exceptionCode = ILLEGAL_DATA_ADDRESS;
// Exception, address does not exist
      } else {
        reply.arybytes[2] = packBits(pDiscretes,
                                     &reply.arybytes[3],
                                     uintSAddr, uintEAddr);
        uintLength = reply.arybytes[2] + 3;
      }
      break;
    case READ_HOLDING_REGISTERS:
    case READ_INPUT_REGISTERS:
    case READ_STATUS_INPUTS:
      if ( request.arybytes[1] == READ_INPUT_REGISTERS ) {
        pRegisters = pInputs;
      } else {
        pRegisters = pHoldingRegs;
      }
      if ( validRegisters(pRegisters, uintSAddr, uintEAddr) == FALSE ) {
// Exception, address does not exist
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      } else {
        reply.arybytes[2] = packRegisters(pRegisters,
                                          &reply.arybytes[3],
                                          uintSAddr, uintEAddr);
        uintLength = reply.arybytes[2] + 3;
      }
      break;
    case FORCE_SINGLE_COIL:
      if ( validDiscretes(pCoils, uintSAddr, uintSAddr) == FALSE ) {
// Exception, address does not exist
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      } else {
        boolean blnOk = FALSE;

        if ( request.arybytes[4] == 0xff &&
             request.arybytes[5] == 0x0 ) {
// Force coil on
          coilState(uintSAddr, TRUE);
          blnOk = TRUE;
        } else if ( request.arybytes[4] == 0x0 &&
                    request.arybytes[5] == 0x0 ) {
// Force coil off
          coilState(uintSAddr, FALSE);
          blnOk = TRUE;
        }
        if ( blnOk == TRUE ) {
// Send back response
          reply.arybytes[2] = request.arybytes[2];
          reply.arybytes[3] = request.arybytes[3];
          reply.arybytes[4] = request.arybytes[4];
          reply.arybytes[5] = request.arybytes[5];
        }
      }
      break;
    case PRESET_SINGLE_REGISTER:
      if ( validRegisters(pHoldingRegs, uintSAddr, uintSAddr) == FALSE ) {
// Exception, address does not exist
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      } else {
        setRegister(uintSAddr, request.arybytes[4], request.arybytes[5]);
        reply.arybytes[2] = request.arybytes[2];
        reply.arybytes[3] = request.arybytes[3];
        reply.arybytes[4] = request.arybytes[4];
        reply.arybytes[5] = request.arybytes[5];
      }
      break;
    case FORCE_MULTIPLE_COILS:
      if ( uintItemCount > MAX_DISCRETES_IN_FC15 ||
           uintItemCount > (request.arybytes[6] * 8) ) {
// Exception, either to many items or the coil count doesn't match byte count
        exceptionCode = ILLEGAL_DATA_ADDRESS;
        break;
      }
    case PRESET_MULTIPLE_REGISTERS:
      if ( (request.arybytes[1] == PRESET_MULTIPLE_REGISTERS &&
            uintItemCount > MAX_REGISTERS_IN_FC16) ) {
// Exception, either to many items or the coil count doesn't match byte count
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      } else if ( validDiscretes(pCoils,
                                 uintSAddr, uintEAddr) == FALSE ) {
        reply.arybytes[1] |= EXCEPTION_FLAG;
      } else {
        uintOffset = 0;
        uintAddr = uintSAddr;

        while( uintAddr<=uintEAddr ) {
          if ( request.arybytes[1] == PRESET_MULTIPLE_REGISTERS ) {
            setRegister(uintAddr,
                        request.arybytes[7 + uintOffset],
                        request.arybytes[8 + uintOffset]);
            uintOffset += 2;
          } else {
            for(uintBit=0; uintBit<8; uintBit++, uintAddr++ ) {
              coilState(uintAddr,
                        request.arybytes[7 + uintOffset] & (1 << uintBit));
            }
            uintOffset++;
          }
          uintAddr++;
        }
        reply.arybytes[2] = request.arybytes[2];
        reply.arybytes[3] = request.arybytes[3];
        reply.arybytes[4] = request.arybytes[4];
        reply.arybytes[5] = request.arybytes[5];
      }
      break;
    }
    if( exceptionCode != NO_EXCEPTION ) {
      reply.arybytes[1] |= EXCEPTION_FLAG;
      reply.arybytes[2] = (byte)exceptionCode;
      exceptionCode = NO_EXCEPTION;
      uintLength = 3;
    }
    if ( uintLength > 0 ) {
// Calculate CRC for response
      lngCRC = calcCRC(reply.arybytes, uintLength);
      reply.arybytes[uintLength] = (lngCRC & 0xff);
      uintLength++;
      reply.arybytes[uintLength] = (lngCRC >> 0x08);
      uintLength++;
// Write the response
      for( bytIndex=0; bytIndex<uintLength; bytIndex++ ) {
        UART1_Write(reply.arybytes[bytIndex]);
      }
// Wait for all data to be transmitted
      while( !TXSTA.TRMT ) ;
    }
  }
// Reset, ready for next receipt
  request.arybytes[0] = 0;
  blnCheckBuffer = FALSE;
// Reset the buffer index
  bytIndex = 0;
}
/**
 * Function:
 * modbusSlave
 *
 * Parameters:
 *  eBaud, see modbusSlave.h for options
 *  bytAddress, the modbus slave address 1 to 255
 *
 * Returns:
 *  0 if ok, -1 if error
 */
int modbusSlave(baudRate eBaud, const byte bytAddress) {
  if ( bytAddress < 1 ) {
    return -1;
  }
// Make sure the received array is cleared
  memset(&request, 0, sizeof(request));
// Set the slave address
  bytSlaveAddress = bytAddress;
// Initialise the Rx Gap counter
  uintRxGap = 0;
// Ensure there is no exception code
  exceptionCode = NO_EXCEPTION;
// Initialize UART module for the specified bps
  switch( eBaud ) {
  case BAUD_1200:
    UART1_Init(BAUD_1200 * 100);
    uintRxGapSetPt = GAP_MULTIPLIER_AT_1200_BAUD;
    break;
  case BAUD_2400:
    UART1_Init(BAUD_2400 * 100);
    uintRxGapSetPt = GAP_MULTIPLIER_AT_2400_BAUD;
    break;
  case BAUD_4800:
    UART1_Init(BAUD_4800 * 100);
    uintRxGapSetPt = GAP_MULTIPLIER_AT_4800_BAUD;
    break;
  case BAUD_9600:
    UART1_Init(BAUD_9600 * 100);
    uintRxGapSetPt = GAP_MULTIPLIER_AT_9600_BAUD;
    break;
  case BAUD_19200:
    UART1_Init(BAUD_19200 * 100);
    uintRxGapSetPt = GAP_MULTIPLIER_AT_19200_BAUD;
    break;
 case BAUD_38400:
    UART1_Init(BAUD_38400 * 100);
    uintRxGapSetPt = GAP_MULTIPLIER_AT_38400_BAUD;
    break;
  case BAUD_57600:
    UART1_Init(BAUD_57600 * 100);
    uintRxGapSetPt = GAP_MULTIPLIER_AT_57600_BAUD;
    break;
  case BAUD_115200:
    UART1_Init(BAUD_115200 * 100);
    uintRxGapSetPt = GAP_MULTIPLIER_AT_115200_BAUD;
    break;
  default:
    return -1;
  }
// Wait for UART module to stabilize
  Delay_ms(100);
// Timer0 Registers:
// 16-Bit Mode
// Prescaler=1:1
// TMRH Preset=FC
// TMRL Preset=E0
// Freq=20,000.00Hz
// Period=0.05 ms
  T0CON.TMR0ON = 1;// Timer0 On/Off Control bit 1=Enables / 0=Stops
  T0CON.T08BIT = 0;// Timer0 8-bit/16-bit Control bit: 1=8-bit/ 0=16-bit
  T0CON.T0CS   = 0;// TMR0 Clock Source Select bit: 0=Internal
  T0CON.T0SE   = 0;// TMR0 Source Edge Select bit: 0=low/high
  T0CON.PSA    = 1;// Prescaler Assignment bit: 1=NOT assigned/bypassed
  T0CON.T0PS2  = 0;// bits 2-0  PS2:PS0: Prescaler Select bits
  T0CON.T0PS1  = 0;
  T0CON.T0PS0  = 0;
  TMR0H = TIMER0_PRESET_MSB;
  TMR0L = TIMER0_PRESET_LSB;
// Turn ON interrupt on UART1 receive
  RC1IE_bit  = 1;
// Clear interrupt flag
  RC1IF_bit  = 0;
// Enable peripheral interrupts
  PEIE_bit   = 1;
// Enable GLOBAL interrupts
  GIE_bit    = 1;
  return 0;
}